<?php

/**
  * Auto Increment Fix
  *
  * Restores AUTO_INCREMENT values in psa database
  * Refer to https://support.plesk.com/hc/en-us/articles/115003292185
  *
  * Targets: Plesk 12.0 and newer
  * Requires: PHP 5.4+, PHP 7.0+
  * Maintainer: Aleksandr Bashurov
  *
  * Copyright 1999-2019. Plesk International GmbH. All rights reserved.
  */

class Table
{
    public $tableName = null;
    public $aiValue = 0;
    public $maxValue = 0;
    public $pkName = null;

    public function toString()
    {
        return 'Table name:     ' . $this->tableName . PHP_EOL
             . 'AUTO_INCREMENT: ' . $this->aiValue . PHP_EOL
             . 'Max value:      ' . $this->maxValue . PHP_EOL
             . 'Primary key:    ' . $this->pkName . PHP_EOL
             . '----------------' . PHP_EOL;
    }
}

final class Database
{
    private static $connection = null;
    private static $tables = [];

    public function __construct(\Environment $probe)
    {
        $password = $probe->getAdminPassword();
        $mysql = mysqli_init();
        if (!$mysql) {
            throw new \Exception('Failed to init MySQLi instance!');
        }
        if (!$mysql->options(MYSQLI_OPT_CONNECT_TIMEOUT, 10)) {
            throw new \Exception('Failed to configure connection timeout!');
        }
        /* @ here is for Plesk with upgraded MySQL/MariaDB instances */
        if (! @$mysql->real_connect(
            'p:localhost',
            'admin',
            $probe->getAdminPassword(),
            'psa',
            $probe->getMysqlPort()
        )) {
            throw new \Exception('Connection to Plesk database failed: ' . mysqli_connect_error());
        }
        if (!$mysql->set_charset('UTF8')) {
            throw new \Exception('Failed to set charset to UTF8: ' . $mysql->error);
        }
        static::$connection = $mysql;
        Logger::debug('Established database connection');
        $this->getApplicableTables();
    }

    private function getApplicableTables()
    {
        $tables = [];
        $query = static::$connection->prepare(
            'SELECT TABLE_NAME, AUTO_INCREMENT
             FROM INFORMATION_SCHEMA.TABLES
             WHERE TABLE_SCHEMA = "psa"
             AND AUTO_INCREMENT IS NOT NULL'
        );
        $query->execute();
        if (!$query) {
            throw new \Exception(
                'Could not extract list of tables with increments: '
                . static::$connection->error
            );
        }
        $query->bind_result($table_name, $ai_value);
        while ($row = $query->fetch()) {
            Logger::info('Querying INFORMATION_SCHEMA for ' . $table_name);
            $table = new Table;
            $table->tableName = $table_name;
            $table->aiValue = $ai_value;
            $tables[] = $table;
        }
        $query->close();
        foreach ($tables as $table) {
            $this->getPrimaryKey($table);
            $this->getMaxValue($table);
            Logger::debug(
                'Table ' . $table->tableName
                . ' (Primary Key = ' . $table->pkName
                . ', Auto Increment = ' . $table->aiValue
                . ', Max Value = ' . $table->maxValue . ')'
            );
            if ($table->maxValue >= $table->aiValue) {
                Logger::debug(
                    'Table ' . $table->tableName . ' has improper AI value'
                );
                static::$tables[] = $table;
            }
        }
    }

    private function updateIncrement(\Table $table)
    {
        $query = static::$connection->prepare(
            'ALTER TABLE ' . static::$connection->real_escape_string($table->tableName)
            . ' AUTO_INCREMENT = ' . static::$connection->real_escape_string($table->maxValue + 1)
        );
        $query->execute();
        if (!$query) {
            throw new \Exception(
                'Could not update auto increment of table '
                . $table->tableName
                . ': '
                . static::$connection->error
            );
        }
        $query->close();
    }

    private function getMaxValue(\Table $table)
    {
        $query = static::$connection->prepare(
            'SELECT MAX(' . static::$connection->real_escape_string($table->pkName) . ')
             FROM ' . static::$connection->real_escape_string($table->tableName)
        );
        $query->execute();
        if (!$query) {
            throw new \Exception(
                'Could not extract max value of table '
                . $table->tableName
                . ' primary key: '
                . static::$connection->error
            );
        }
        $query->bind_result($max_value);
        $query->fetch();
        $table->maxValue = $max_value;
        $query->close();
    }

    private function getPrimaryKey(\Table $table)
    {
        $query = static::$connection->prepare(
            'SELECT COLUMN_NAME
             FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
             WHERE TABLE_SCHEMA = "psa"
             AND CONSTRAINT_NAME = "PRIMARY"
             AND TABLE_NAME = ?'
        );
        $query->bind_param("s", $table->tableName);
        $query->execute();
        if (!$query) {
            throw new \Exception(
                'Could not extract primary key of table '
                . $table->tableName
                . ': '
                . static::$connection->error
            );
        }
        $query->bind_result($primary_key);
        $query->fetch();
        $table->pkName = $primary_key;
        $query->close();
    }

    public function restoreValues()
    {
        foreach (static::$tables as $table) {
            if ($table->aiValue <= $table->maxValue) {
                Logger::write('Fixing table ' . $table->tableName);
                $this->updateIncrement($table);
                Logger::write('Completed');
            }
        }
    }

    public function getAffectedTables()
    {
        foreach (static::$tables as $table) {
            if ($table->aiValue <= $table->maxValue) {
                Logger::write($table->toString());
            }
        }
    }

    public function getCountTables()
    {
        return count(static::$tables);
    }

}

final class Environment
{
    const OS_WINDOWS = 1;
    const OS_LINUX = 2;

    private static $password = null;
    private static $version = null;
    private static $port = null;
    private static $os = null;

    public function getAdminPassword()
    {
        $version = $this->getPleskVersion();
        if (static::$password === null) {
            if ($this->getOs() === static::OS_WINDOWS) {
                if ($version >= 17800) {
                    exec('plesk sbin psadb -g', $password, $rval);
                    if ($rval) {
                        throw new \Exception('Failed to extract Plesk database password!');
                    }
                    static::$password = $password[0];
                } else {
                    if ($version >= 12500) {
                        exec('plesk sbin plesksrvclient -get -nogui', $password, $rval);
                    } else {
                        exec(
                            '"' . $this->getRootDir() . '\admin\bin\plesksrvclient.exe" -get -nogui',
                            $password,
                            $rval
                        );
                    }
                    if ($rval) {
                        throw new Exception('Failed to extract Plesk database password!');
                    }
                    static::$password = substr($password[0], 22);
                }
            } else {
                static::$password = trim(file_get_contents('/etc/psa/.psa.shadow'));
            }
        }
        return static::$password;
    }

    public function getMysqlPort()
    {
        if (static::$port === null) {
            if ($this->getOs() === static::OS_WINDOWS) {
                static::$port = 8306;
            } else {
                static::$port = 3306;
            }
            Logger::debug('Using MySQL port ' . static::$port);
        }
        return static::$port;
    }

    private function getOs()
    {
        if (static::$os === null) {
            if (strtoupper(substr(PHP_OS, 0, 3)) === 'WIN') {
                static::$os = static::OS_WINDOWS;
            } else {
                static::$os = static::OS_LINUX;
            }
            Logger::debug('OS: ' . (static::$os === static::OS_WINDOWS ? 'Windows' : 'Linux'));
        }
        return static::$os;
    }

    private function getPleskVersion()
    {
        if (static::$version === null) {
            static::$version = intval(filter_var(
                explode(' ', file_get_contents($this->getRootDir() . 'version'))[0],
                FILTER_SANITIZE_NUMBER_INT
            ));
            Logger::debug('Plesk version: ' . static::$version);
        }
        return static::$version;
    }

    private function getRootDir()
    {
        if ($this->getOs() === static::OS_WINDOWS) {
            $envdir = getenv('plesk_dir');
            if ($envdir === false || !is_readable($envdir)) {
                throw new \Exception('Could not find Plesk installation directory!');
            }
            return $envdir;
        } else {
            return '/usr/local/psa/';
        }
    }

}

final class Configuration
{
    private static $verbosity = 0;
    private static $fix = false;

    /**
     * @param int $new_level
     */
    public static function setVerbosity($new_level)
    {
        if (!is_int($new_level)) {
            throw new \Exception("Type Error: expected integer, received " . gettype($new_level));
        }
        static::$verbosity = $new_level;
        return new static();
    }

    public static function getVerbosity()
    {
        return static::$verbosity;
    }

    /**
     * @param int $new_value
     */
    public static function setAutoFix($new_value)
    {
        if (!is_bool($new_value)) {
            throw new \Exception("Type Error: expected boolean, received " . gettype($new_value));
        }
        static::$fix = $new_value;
        return new static();
    }

    public static function getAutoFix()
    {
        return static::$fix;
    }

}

final class Logger
{
    private function __construct() {}

    public static function debug($value)
    {
        if (Configuration::getVerbosity() >= 2) {
            fwrite(STDERR, 'DEBUG: ' . $value . PHP_EOL);
        }
    }

    public static function info($value)
    {
        if (Configuration::getVerbosity() >= 1) {
            fwrite(STDOUT, 'INFO: ' . $value . PHP_EOL);
        }
    }

    public static function error($value)
    {
        fwrite(STDERR, '!ERROR: ' . $value . PHP_EOL);
    }

    public static function write($value)
    {
        fwrite(STDOUT, $value . PHP_EOL);
    }
}

function helpCommand()
{
    $script_name = basename(__FILE__);
    Logger::write(<<<HELP
$script_name: Fix AUTO_INCREMENT values for applicable tables
Version: 1.0

Usage:
   $script_name -y -v -d

      Options:
   -v     Enables verbose output

   -d     Enables debug output

   -y     Automatically fixes incorrect AUTO_INCREMENT values
          By default, the misconfiguration is only reported to user

   -h     Shows this help message
HELP
   );
}

function saneOpts()
{
    /* Thank you for left-associative ternary operators, Zend */
    $options = getopt('yvdh');
    if ($options === null) {
        return [];
    }
    $sane_options = [
        'autoFix' => array_key_exists('y', $options),
        'verbosityLevel' => (
            array_key_exists('d', $options) ? 2 : (
            array_key_exists('v', $options) ? 1 : (
            0
        ))),
        'helpMessage' => array_key_exists('h', $options)
    ];
    return $sane_options;
}

function main()
{
    $script_name = basename(__FILE__);
    $options = saneOpts();
    if ($options['helpMessage']) {
        helpCommand();
        exit(0);
    }
    Configuration::setAutoFix($options['autoFix'])
        ->setVerbosity($options['verbosityLevel']);
    Logger::write('Analyzing tables in `psa` database');
    $env = new Environment();
    $db_manager = new Database($env);
    if ($db_manager->getCountTables() > 0) {
        Logger::write('Following tables are affected: ' . PHP_EOL);
        $db_manager->getAffectedTables();
        if (Configuration::getAutoFix()) {
            $db_manager->restoreValues();
        } else {
            Logger::error(
                'No fix actions are to be performed' . PHP_EOL
                . 'If you want to fix inconsistencies, run the script with -y option: ' . $script_name . ' -y'
            );
        }
    } else {
        Logger::write('No inconsistencies were detected.');
    }
}

main();

